@@ -5,6 +5,29 @@ from enum import Enum, auto
 from typing import List, Optional, Union
 from dataclasses import dataclass
 
+consistent_datatypes = {"numeric": ["bit", "smallmoney", "money", "int",
+                                    "decimal", "decimal(4,2)", "decimal(3,2)", "float", "float64", "double",
+                                    "bigdecimal", "int32", "int64", "bigint",
+                                    "numeric", "bignumeric", "smallint", "integer",
+                                    "tinyint", "byteint", "int8", "long",
+                                    "float8", "hugeint", "int4", "signed",
+                                    "real", "float4", "int2", "short",
+                                    "int1", "ubigint", "uinteger", "usmallint",
+                                    "utinyint", "bool", "boolean", "mediumint",
+                                    "double precision", "dec", "smallserial", "serial",
+                                    "bigserial", "number", "number(3,2)", "binary", "smalldecimal"],
+                        "date":
+                            ["date", "datetimeoffset", "datetime2", "smalldatetime",
+                             "timestamp", "datetime", "time", "year",
+                             "interval", "timestamp_ltz", "timestamp_ntz", "timestamp_tz"],
+                        "string":
+                            ["string", "varchar", "varchar2", "char",
+                             "bpchar", "text", "nchar", "nvarchar",
+                             "tinyblob", "blob", "shorttext", "clob",
+                             "nclob", "ntext", "tinytext", "mediumtext",
+                             "longtext", "mediumblob", "longblob", "image",
+                             "character", "character varying", "binary", "varbinary"]}
+
 
 class SQLFunction(Enum):
     MAX = auto()
@@ -221,6 +244,11 @@ class Join:
     right_on: Optional[List[str]] = None
 
 
+@dataclass
+class GroupBy:
+    field: Field
+
+
 @dataclass
 class Query:
     fields: List[Field]
@@ -229,32 +257,10 @@ class Query:
     id: Optional[str] = None
     datasets: Optional[List[From]] = None
     where: Optional[List[Union[Condition, CompositeCondition]]] = None
+    groupBy: Optional[List[GroupBy]] = None
     orderBy: Optional[List[Sorting]] = None
     limit: Optional[Union[int, str]] = None
     offset: Optional[Union[int, str]] = None
-    consistent_datatypes = {"numeric":
-                                ["bit", "smallmoney", "money", "int",
-                                 "decimal", "decimal(4,2)", "decimal(3,2)", "float", "float64", "double",
-                                 "bigdecimal", "int32", "int64", "bigint",
-                                 "numeric", "bignumeric", "smallint", "integer",
-                                 "tinyint", "byteint", "int8", "long",
-                                 "float8", "hugeint", "int4", "signed",
-                                 "real", "float4", "int2", "short",
-                                 "int1", "ubigint", "uinteger", "usmallint",
-                                 "utinyint", "bool", "boolean", "mediumint",
-                                 "double precision", "dec", "smallserial", "serial",
-                                 "bigserial", "number", "number(3,2)", "binary", "smalldecimal"],
-                            "date":
-                                ["date", "datetimeoffset", "datetime2", "smalldatetime",
-                                 "timestamp", "datetime", "time", "year",
-                                 "interval", "timestamp_ltz", "timestamp_ntz", "timestamp_tz"],
-                            "string":
-                                ["string", "varchar", "varchar2", "char",
-                                 "bpchar", "text", "nchar", "nvarchar",
-                                 "tinyblob", "blob", "shorttext", "clob",
-                                 "nclob", "ntext", "tinytext", "mediumtext",
-                                 "longtext", "mediumblob", "longblob", "image",
-                                 "character", "character varying", "binary", "varbinary"]}
 
     def to_sql(self, dataset: str = None):
         sql = "SELECT {} FROM {}"
@@ -317,7 +323,81 @@ class Query:
 
         return sql
 
+    def _field_spell_out(self, field):
+        # otherwise, check if internal and original data types are not consistent each other
+        global consistent_datatypes
+        consistent_vec = consistent_datatypes[field.internalDataType.lower()]
+        if (
+                field.internalDataType.lower() == "numeric" or field.internalDataType.lower() == "string") and field.sourceDataType.lower() not in consistent_vec:
+            # if internal and original data types are not consistent each other use this pattern
+            # aggregation column (original data type, to internal data type ) AS Alias
+            if field.alias is None:
+                if field.aggregation is not None and isinstance(field.aggregation, str):
+                    string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ")" + '"'
+                elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
+                    string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ")" + '"'
+                else:
+                    string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ")" + '"'
+            else:
+                if field.aggregation is not None and isinstance(field.aggregation, str):
+                    string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ") AS " + field.alias + '"'
+                elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
+                    string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ") AS " + field.alias + '"'
+                else:
+                    string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ") AS " + field.alias + '"'
+        elif field.internalDataType.lower() == "date" and field.sourceDataType.lower() not in consistent_vec:
+            # if internal and original data types are not consistent each other use this pattern
+            # aggregation column (original data type, to internal data type, date_format YYYY-MM-DD ) AS Alias
+            if field.sourceFormat is not None:
+                if field.alias is None:
+                    if field.aggregation is not None and isinstance(field.aggregation, str):
+                        string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ")" + '"'
+                    elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
+                        string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ")" + '"'
+                    else:
+                        string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ")" + '"'
+                else:
+                    if field.aggregation is not None and isinstance(field.aggregation, str):
+                        string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ") AS " + field.alias + '"'
+                    elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
+                        string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ") AS " + field.alias + '"'
+                    else:
+                        string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ") AS " + field.alias + '"'
+            else:
+                if field.alias is None:
+                    if field.aggregation is not None and isinstance(field.aggregation, str):
+                        string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ")" + '"'
+                    elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
+                        string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ")" + '"'
+                    else:
+                        string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ")" + '"'
+                else:
+                    if field.aggregation is not None and isinstance(field.aggregation, str):
+                        string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ") AS " + field.alias + '"'
+                    elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
+                        string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ") AS " + field.alias + '"'
+                    else:
+                        string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ") AS " + field.alias + '"'
+        else:
+            # else
+            if field.alias is None:
+                if field.aggregation is not None and isinstance(field.aggregation, str):
+                    string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ")" + '"'
+                elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
+                    string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ")" + '"'
+                else:
+                    string = '"' + field.column + " (" + field.sourceDataType + ")" + '"'
+            else:
+                if field.aggregation is not None and isinstance(field.aggregation, str):
+                    string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ") AS " + field.alias + '"'
+                elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
+                    string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ") AS " + field.alias + '"'
+                else:
+                    string = '"' + field.column + " (" + field.sourceDataType + ") AS " + field.alias + '"'
+        return string
+
     def _where_spell_out(self, where):
+        global consistent_datatypes
         conditions = []
 
         for where_condition in where:
@@ -348,7 +428,7 @@ class Query:
                 else:
                     operator = None
 
-                consistent_vec = self.consistent_datatypes[where_condition.field.internalDataType.lower()]
+                consistent_vec = consistent_datatypes[where_condition.field.internalDataType.lower()]
                 if (
                         where_condition.field.internalDataType.lower() == "numeric" or where_condition.field.internalDataType.lower() == "string") \
                         and where_condition.field.sourceDataType.lower() not in consistent_vec:
@@ -397,75 +477,7 @@ class Query:
                     else:
                         string = '"' + field.sqlExpression + " AS " + field.alias + '"'
             else:
-                # otherwise, check if internal and original data types are not consistent each other
-                consistent_vec = self.consistent_datatypes[field.internalDataType.lower()]
-                if (
-                        field.internalDataType.lower() == "numeric" or field.internalDataType.lower() == "string") and field.sourceDataType.lower() not in consistent_vec:
-                    # if internal and original data types are not consistent each other use this pattern
-                    # aggregation column (original data type, to internal data type ) AS Alias
-                    if field.alias is None:
-                        if field.aggregation is not None and isinstance(field.aggregation, str):
-                            string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ")" + '"'
-                        elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
-                            string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ")" + '"'
-                        else:
-                            string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ")" + '"'
-                    else:
-                        if field.aggregation is not None and isinstance(field.aggregation, str):
-                            string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ") AS " + field.alias + '"'
-                        elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
-                            string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ") AS " + field.alias + '"'
-                        else:
-                            string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ") AS " + field.alias + '"'
-                elif field.internalDataType.lower() == "date" and field.sourceDataType.lower() not in consistent_vec:
-                    # if internal and original data types are not consistent each other use this pattern
-                    # aggregation column (original data type, to internal data type, date_format YYYY-MM-DD ) AS Alias
-                    if field.sourceFormat is not None:
-                        if field.alias is None:
-                            if field.aggregation is not None and isinstance(field.aggregation, str):
-                                string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ")" + '"'
-                            elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
-                                string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ")" + '"'
-                            else:
-                                string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ")" + '"'
-                        else:
-                            if field.aggregation is not None and isinstance(field.aggregation, str):
-                                string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ") AS " + field.alias + '"'
-                            elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
-                                string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ") AS " + field.alias + '"'
-                            else:
-                                string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format " + field.sourceFormat + ") AS " + field.alias + '"'
-                    else:
-                        if field.alias is None:
-                            if field.aggregation is not None and isinstance(field.aggregation, str):
-                                string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ")" + '"'
-                            elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
-                                string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ")" + '"'
-                            else:
-                                string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ")" + '"'
-                        else:
-                            if field.aggregation is not None and isinstance(field.aggregation, str):
-                                string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ") AS " + field.alias + '"'
-                            elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
-                                string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ") AS " + field.alias + '"'
-                            else:
-                                string = '"' + field.column + " (" + field.sourceDataType + ", to " + field.internalDataType + ", format YYYY-MM-DD" + ") AS " + field.alias + '"'
-                else:
-                    # else
-                    if field.alias is None:
-                        if field.aggregation is not None and isinstance(field.aggregation, str):
-                            string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ")" + '"'
-                        elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
-                            string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ")" + '"'
-                        else:
-                            string = '"' + field.column + " (" + field.sourceDataType + ")" + '"'
-                    else:
-                        if field.aggregation is not None and isinstance(field.aggregation, str):
-                            string = '"' + field.aggregation + " " + field.column + " (" + field.sourceDataType + ") AS " + field.alias + '"'
-                        elif field.aggregation is not None and isinstance(field.aggregation, SQLFunction):
-                            string = '"' + field.aggregation.name + " " + field.column + " (" + field.sourceDataType + ") AS " + field.alias + '"'
-                        else:
-                            string = '"' + field.column + " (" + field.sourceDataType + ") AS " + field.alias + '"'
+                string = self._field_spell_out(field)
             fields.append(string)
         spell_str += ", ".join(fields) + "]"
 
@@ -478,6 +490,21 @@ class Query:
             condition_str = ['"' + cond + '"' for cond in condition_str]
             spell_str += ", ".join(condition_str) + "]"
 
+        if self.groupBy is not None and self.groupBy:
+            spell_str += ', "group_by":['
+            groupby_vec = []
+            for clause in self.groupBy:
+                new_field = Field(column=clause.field.column,
+                                  alias=None,
+                                  aggregation=clause.field.aggregation,
+                                  sourceDataType=clause.field.sourceDataType,
+                                  internalDataType=clause.field.internalDataType,
+                                  sourceFormat=clause.field.sourceFormat,
+                                  sqlExpression=clause.field.sqlExpression)
+                string = self._field_spell_out(new_field)
+                groupby_vec.append(string)
+            spell_str += ", ".join(groupby_vec) + "]"
+
         if self.orderBy is not None and self.orderBy:
             spell_str += ', "order_by":['
             orderby_vec = []
@@ -513,13 +540,6 @@ class SmartQuery:
             spell_array.append(spell)
         return spell_array
 
-    def spell_out_untyped(self):
-        spell_array = []
-        for query in self.queries:
-            spell = query.spell_out_untyped()
-            spell_array.append(spell)
-        return spell_array
-
     @staticmethod
     def _get_tokens():
         f = open(os.path.join(os.path.dirname(__file__), "askdata_config", "compression_tokens.json"))
